from io import StringIO
import traceback
import sys
import functools
import json
import socket
import os
import asyncio
import time
import threading
from urllib.parse import urlparse, urlunparse
from socket import gethostbyname
from aiohttp import ClientSession, __version__


_DNS_CACHE = {}
_STOP = False
_TIMER = None
IS_AIOHTTP2 = __version__[0] == '2'


def get_timer():
    return _TIMER


def set_timer(value=None):
    global _TIMER
    if value is None:
        value = int(time.time())
    _TIMER = value


def stop():
    global _STOP
    _STOP = True


def is_stopped():
    return _STOP


def resolve(url):
    parts = urlparse(url)

    if '@' in parts.netloc:
        username, password = parts.username, parts.password
        netloc = parts.netloc.split('@', 1)[1]
    else:
        username, password = None, None
        netloc = parts.netloc

    if ':' in netloc:
        host = netloc.split(':')[0]
    else:
        host = netloc

    port_provided = False
    if not parts.port and parts.scheme == 'https':
        port = 443
    elif not parts.port and parts.scheme == 'http':
        port = 80
    else:
        port = parts.port
        port_provided = True

    original = host
    resolved = None
    if host in _DNS_CACHE:
        resolved = _DNS_CACHE[host]
    else:
        try:
            resolved = gethostbyname(host)
            _DNS_CACHE[host] = resolved
        except socket.gaierror:
            return url, original, host

    # Don't use a resolved hostname for SSL requests otherwise the
    # certificate will not match the IP address (resolved)
    host = resolved if parts.scheme != 'https' else host
    netloc = host
    if port_provided:
        netloc += ':%d' % port
    if username is not None:
        if password is not None:
            netloc = '%s:%s@%s' % (username, password, netloc)
        else:
            netloc = '%s@%s' % (username, netloc)

    if port not in (443, 80):
        host += ':%d' % port
        original += ':%d' % port

    new = urlunparse((parts.scheme, netloc, parts.path or '', '',
                      parts.query or '', parts.fragment or ''))
    return new, original, host


class OptionError(Exception):
    pass


def _expand_args(args, options):
    for key, val in options.items():
        setattr(args, key, val)


def expand_options(config, scenario, args):
    if not isinstance(config, str):
        try:
            config = json.loads(config.read())
        except Exception:
            raise OptionError("Can't parse %r" % config)
    else:
        if not os.path.exists(config):
            raise OptionError("Can't find %r" % config)

        with open(config) as f:
            try:
                config = json.loads(f.read())
            except ValueError:
                raise OptionError("Can't parse %r" % config)

    if 'molotov' not in config:
        raise OptionError("Bad config -- no molotov key")

    if 'tests' not in config['molotov']:
        raise OptionError("Bad config -- no molotov/tests key")

    if scenario not in config['molotov']['tests']:
        raise OptionError("Can't find %r in the config" % scenario)

    _expand_args(args, config['molotov']['tests'][scenario])


def _run_in_fresh_loop(coro, timeout=30):
    thres = []
    thexc = []

    def run():
        loop = asyncio.new_event_loop()
        try:
            task = loop.create_task(coro(loop=loop))
            thres.append(loop.run_until_complete(task))
        except Exception as e:
            thexc.append(e)
        finally:
            loop.close()

    th = threading.Thread(target=run)
    th.start()
    th.join(timeout=timeout)

    # re-raise a thread exception
    if len(thexc) > 0:
        raise thexc[0]

    return thres[0]


async def _request(endpoint, verb='GET', session_options=None,
                   json=False, loop=None, **options):
    if session_options is None:
        session_options = {}

    async with ClientSession(loop=loop, **session_options) as session:
        meth = getattr(session, verb.lower())
        result = {}
        async with meth(endpoint, **options) as resp:
            if json:
                result['content'] = await resp.json()
            else:
                result['content'] = await resp.text()
            result['status'] = resp.status
            result['headers'] = resp.headers

        return result


def request(endpoint, verb='GET', session_options=None, **options):
    """Performs a synchronous request.

    Uses a dedicated event loop and aiohttp.ClientSession object.

    Options:

    - endpoint: the endpoint to call
    - verb: the HTTP verb to use (defaults: GET)
    - session_options: a dict containing options to initialize the session
      (defaults: None)
    - options: extra options for the request (defaults: None)

    Returns a dict object with the following keys:

    - content: the content of the response
    - status: the status
    - headers: a dict with all the response headers
    """
    req = functools.partial(_request, endpoint, verb, session_options,
                            **options)
    return _run_in_fresh_loop(req)


def json_request(endpoint, verb='GET', session_options=None, **options):
    """Like :func:`molotov.request` but extracts json from the response.
    """
    req = functools.partial(_request, endpoint, verb, session_options,
                            json=True, **options)
    return _run_in_fresh_loop(req)


_VARS = {}


def set_var(name, value):
    """Sets a global variable.

    Options:

    - name: name of the variable
    - value: object to set
    """
    _VARS[name] = value


def get_var(name, factory=None):
    """Gets a global variable given its name.

    If factory is not None and the variable is not set, factory
    is a callable that will set the variable.

    If not set, returns None.
    """
    if name not in _VARS and factory is not None:
        _VARS[name] = factory()
    return _VARS.get(name)


# taken from https://stackoverflow.com/a/37211337
def _make_sleep():
    async def sleep(delay, result=None, *, loop=None):
        coro = asyncio.sleep(delay, result=result, loop=loop)
        task = asyncio.ensure_future(coro, loop=loop)
        sleep.tasks.add(task)
        try:
            return await task
        except asyncio.CancelledError:
            return result
        finally:
            sleep.tasks.remove(task)

    sleep.tasks = set()
    sleep.cancel_all = lambda: sum(task.cancel() for task in sleep.tasks)
    return sleep


cancellable_sleep = _make_sleep()


def printable_error(error, tb=None):
    printable = [repr(error)]
    if tb is None:
        tb = sys.exc_info()[2]
    printed = StringIO()
    traceback.print_tb(tb, file=printed)
    printed.seek(0)
    for line in printed.readlines():
        printable.append(line.rstrip('\n'))
    return printable



import time
import random
import os
import signal
import asyncio
from unittest.mock import patch

from molotov.api import scenario, global_setup
from molotov.tests.support import (TestLoop, coserver, dedicatedloop, set_args,
                                   skip_pypy, only_pypy, catch_sleep,
                                   dedicatedloop_noclose)
from molotov.tests.statsd import UDPServer
from molotov.run import run, main
from molotov.sharedcounter import SharedCounters
from molotov.util import request, json_request, set_timer
from molotov import __version__


_HERE = os.path.dirname(__file__)
_CONFIG = os.path.join(_HERE, 'molotov.json')
_RES = []
_RES2 = {}


class TestRunner(TestLoop):
    def setUp(self):
        super(TestRunner, self).setUp()
        _RES[:] = []
        _RES2.clear()

    def _get_args(self):
        args = self.get_args()
        args.statsd = True
        args.statsd_address = 'udp://127.0.0.1:9999'
        args.scenario = 'molotov.tests.test_run'
        return args

    @dedicatedloop_noclose
    def test_redirect(self):
        @scenario(weight=10)
        async def _one(session):
            # redirected
            async with session.get('http://localhost:8888/redirect') as resp:
                redirect = resp.history
                assert redirect[0].status == 302
                assert resp.status == 200

            # not redirected
            async with session.get('http://localhost:8888/redirect',
                                   allow_redirects=False) as resp:
                redirect = resp.history
                assert len(redirect) == 0
                assert resp.status == 302
                content = await resp.text()
                assert content == ''

            _RES.append(1)

        args = self._get_args()
        args.verbose = 2
        args.max_runs = 2
        with coserver():
            run(args)

        self.assertTrue(len(_RES) > 0)

    @dedicatedloop_noclose
    def test_runner(self):
        test_loop = asyncio.get_event_loop()

        @global_setup()
        def something_sync(args):
            grab = request('http://localhost:8888')
            self.assertEqual(grab['status'], 200)
            grab_json = json_request('http://localhost:8888/molotov.json')
            self.assertTrue('molotov' in grab_json['content'])

        @scenario(weight=10)
        async def here_one(session):
            async with session.get('http://localhost:8888') as resp:
                await resp.text()
            _RES.append(1)

        @scenario(weight=90)
        async def here_two(session):
            if session.statsd is not None:
                session.statsd.incr('yopla')
            _RES.append(2)

        args = self._get_args()
        server = UDPServer('127.0.0.1', 9999, loop=test_loop)
        _stop = asyncio.Future()

        async def stop():
            await _stop
            await server.stop()

        server_task = asyncio.ensure_future(server.run())
        stop_task = asyncio.ensure_future(stop())
        args.max_runs = 3

        with coserver():
            run(args)

        _stop.set_result(True)
        test_loop.run_until_complete(asyncio.gather(server_task, stop_task))

        self.assertTrue(len(_RES) > 0)

        udp = server.flush()
        self.assertTrue(len(udp) > 0)

    @dedicatedloop
    def test_main(self):
        with set_args('molotov', '-cq', '-d', '1', 'molotov/tests/example.py'):
            main()

    def _test_molotov(self, *args):
        if '--duration' not in args and '-d' not in args:
            args = list(args) + ['--duration', '10']
        rc = 0
        with set_args('molotov', *args) as (stdout, stderr):
            try:
                main()
            except SystemExit as e:
                rc = e.code
        return stdout.read().strip(), stderr.read().strip(), rc

    @dedicatedloop
    def test_version(self):
        stdout, stderr, rc = self._test_molotov('--version')
        self.assertEqual(stdout, __version__)

    @dedicatedloop
    def test_empty_scenario(self):
        stdout, stderr, rc = self._test_molotov('')
        self.assertTrue('Cannot import' in stdout)

    @dedicatedloop
    def test_no_scenario(self):
        stdout, stderr, rc = self._test_molotov()
        self.assertTrue('Cannot import' in stdout)

    @dedicatedloop
    def test_config_no_scenario(self):
        stdout, stderr, rc = self._test_molotov('-c', '--config', _CONFIG,
                                                'DONTEXIST')
        wanted = "Can't find 'DONTEXIST' in the config"
        self.assertTrue(wanted in stdout)

    @dedicatedloop
    def test_config_verbose_quiet(self):
        stdout, stderr, rc = self._test_molotov(
            '-qv', '--config', _CONFIG)
        wanted = "You can't"
        self.assertTrue(wanted in stdout)

    @dedicatedloop
    def test_config_no_scenario_found(self):
        stdout, stderr, rc = self._test_molotov(
            '-c', 'molotov.tests.test_run')
        wanted = "No scenario was found"
        self.assertTrue(wanted in stdout)

    @dedicatedloop
    def test_config_no_single_mode_found(self):

        @scenario(weight=10)
        async def not_me(session):
            _RES.append(3)

        stdout, stderr, rc = self._test_molotov('-c', '-s', 'blah',
                                                'molotov.tests.test_run')
        wanted = "Can't find"
        self.assertTrue(wanted in stdout)

    @dedicatedloop
    def test_name(self):

        @scenario(weight=10)
        async def here_three(session):
            _RES.append(3)

        @scenario(weight=30, name='me')
        async def here_four(session):
            _RES.append(4)

        stdout, stderr, rc = self._test_molotov('-cx', '--max-runs', '2', '-s',
                                                'me', 'molotov.tests.test_run')
        wanted = "SUCCESSES: 2"
        self.assertTrue(wanted in stdout)
        self.assertTrue(_RES, [4, 4])

    @dedicatedloop
    def test_single_mode(self):

        @scenario(weight=10)
        async def here_three(session):
            _RES.append(3)

        stdout, stderr, rc = self._test_molotov('-cx', '--max-runs', '2', '-s',
                                                'here_three',
                                                'molotov.tests.test_run')
        wanted = "SUCCESSES: 2"
        self.assertTrue(wanted in stdout)

    @dedicatedloop
    def test_fail_mode_pass(self):

        @scenario(weight=10)
        async def here_three(session):
            _RES.append(3)

        stdout, stderr, rc = self._test_molotov('-cx', '--max-runs', '2',
                                                '--fail', '1',
                                                '-s', 'here_three',
                                                'molotov.tests.test_run')
        wanted = "SUCCESSES: 2"
        self.assertTrue(wanted in stdout)
        self.assertEqual(rc, 0)

    @dedicatedloop
    def test_fail_mode_fail(self):

        @scenario(weight=10)
        async def here_three(session):
            assert False

        stdout, stderr, rc = self._test_molotov('-cx', '--max-runs', '2',
                                                '--fail', '1', '-s',
                                                'here_three',
                                                'molotov.tests.test_run')
        self.assertEqual(rc, 1)

    @only_pypy
    @dedicatedloop
    def test_uvloop_pypy(self):

        @scenario(weight=10)
        async def here_three(session):
            _RES.append(3)

        orig_import = __import__

        def import_mock(name, *args):
            if name == 'uvloop':
                raise ImportError()
            return orig_import(name, *args)

        with patch('builtins.__import__', side_effect=import_mock):
            stdout, stderr, rc = self._test_molotov('-cx', '--max-runs', '2',
                                                    '-s',
                                                    'here_three', '--uvloop',
                                                    'molotov.tests.test_run')
        wanted = "You can't use uvloop"
        self.assertTrue(wanted in stdout)

    @skip_pypy
    @dedicatedloop
    def test_uvloop_import_error(self):

        @scenario(weight=10)
        async def here_three(session):
            _RES.append(3)

        orig_import = __import__

        def import_mock(name, *args):
            if name == 'uvloop':
                raise ImportError()
            return orig_import(name, *args)

        with patch('builtins.__import__', side_effect=import_mock):
            stdout, stderr, rc = self._test_molotov('-cx', '--max-runs', '2',
                                                    '--console-update', '0',
                                                    '-s',
                                                    'here_three', '--uvloop',
                                                    'molotov.tests.test_run')
        wanted = "You need to install uvloop"
        self.assertTrue(wanted in stdout)

    @skip_pypy
    @dedicatedloop
    def test_uvloop(self):

        @scenario(weight=10)
        async def here_three(session):
            _RES.append(3)

        stdout, stderr, rc = self._test_molotov('-cx', '--max-runs', '2', '-s',
                                                'here_three', '--uvloop',
                                                'molotov.tests.test_run')
        wanted = "SUCCESSES: 2"
        self.assertTrue(wanted in stdout, stdout)

    @dedicatedloop
    def test_delay(self):
        with catch_sleep() as delay:
            @scenario(weight=10, delay=.1)
            async def here_three(session):
                _RES.append(3)

            stdout, stderr, rc = self._test_molotov('--delay', '.6',
                                                    '--console-update', '0',
                                                    '-cx', '--max-runs', '2',
                                                    '-s', 'here_three',
                                                    'molotov.tests.test_run')
            wanted = "SUCCESSES: 2"
            self.assertTrue(wanted in stdout, stdout)
            self.assertEqual(delay, [1, .1, 1, .6, 1, .1, 1, .6, 1])

    @dedicatedloop
    def test_rampup(self):
        with catch_sleep() as delay:
            @scenario(weight=10)
            async def here_three(session):
                _RES.append(3)

            stdout, stderr, rc = self._test_molotov('--ramp-up', '10',
                                                    '--workers', '5',
                                                    '--console-update', '0',
                                                    '-cx', '--max-runs', '2',
                                                    '-s', 'here_three',
                                                    'molotov.tests.test_run')
            # workers should start every 2 seconds since
            # we have 5 workers and a ramp-up
            # the first one starts immediatly, then each worker
            # sleeps 2 seconds more.
            delay = [d for d in delay if d != 0]
            self.assertEqual(delay, [1, 2.0, 4.0, 6.0, 8.0, 1, 1])
            wanted = "SUCCESSES: 10"
            self.assertTrue(wanted in stdout, stdout)

    @dedicatedloop
    def test_sizing(self):
        _RES2['fail'] = 0
        _RES2['succ'] = 0

        with catch_sleep():
            @scenario()
            async def sizer(session):
                if random.randint(0, 20) == 1:
                    _RES2['fail'] += 1
                    raise AssertionError()
                else:
                    _RES2['succ'] += 1

            stdout, stderr, rc = self._test_molotov('--sizing',
                                                    '--console-update', '0',
                                                    '--sizing-tolerance', '5',
                                                    '-s', 'sizer',
                                                    'molotov.tests.test_run')

        ratio = float(_RES2['fail']) / float(_RES2['succ']) * 100.
        self.assertTrue(ratio < 15. and ratio >= 5., ratio)

    @dedicatedloop
    def test_sizing_multiprocess(self):
        counters = SharedCounters('OK', 'FAILED')

        with catch_sleep():
            @scenario()
            async def sizer(session):
                if random.randint(0, 10) == 1:
                    counters['FAILED'] += 1
                    raise AssertionError()
                else:
                    counters['OK'] += 1

            with set_args('molotov', '--sizing', '-p', '2',
                          '--sizing-tolerance', '5',
                          '--console-update', '0',
                          '-s', 'sizer',
                          'molotov.tests.test_run') as (stdout, stderr):
                try:
                    main()
                except SystemExit:
                    pass
            stdout, stderr = stdout.read().strip(), stderr.read().strip()

            # stdout, stderr, rc = self._test_molotov()
            ratio = (float(counters['FAILED'].value) /
                     float(counters['OK'].value) * 100.)
            self.assertTrue(ratio >= 5., ratio)

    @dedicatedloop_noclose
    def test_statsd_multiprocess(self):
        test_loop = asyncio.get_event_loop()

        @scenario()
        async def staty(session):
            session.statsd.incr('yopla')

        server = UDPServer('127.0.0.1', 9999, loop=test_loop)
        _stop = asyncio.Future()

        async def stop():
            await _stop
            await server.stop()

        server_task = asyncio.ensure_future(server.run())
        stop_task = asyncio.ensure_future(stop())
        args = self._get_args()
        args.verbose = 2
        args.processes = 2
        args.max_runs = 5
        args.duration = 1000
        args.statsd = True
        args.statsd_address = 'udp://127.0.0.1:9999'
        args.single_mode = 'staty'
        args.scenario = 'molotov.tests.test_run'

        run(args)

        _stop.set_result(True)
        test_loop.run_until_complete(asyncio.gather(server_task, stop_task))
        udp = server.flush()

        incrs = 0
        for line in udp:
            for el in line.split(b'\n'):
                if el.strip() == b'':
                    continue
                incrs += 1

        # two processes making 5 run each
        self.assertEqual(incrs, 10)

    @dedicatedloop
    def test_timed_sizing(self):
        _RES2['fail'] = 0
        _RES2['succ'] = 0
        _RES2['messed'] = False

        with catch_sleep():
            @scenario()
            async def sizer(session):
                if session.worker_id == 200 and not _RES2['messed']:
                    # worker 2 will mess with the timer
                    # since we're faking all timers, the current
                    # time in the test is always around 0
                    # so to have now() - get_timer() > 60
                    # we need to set a negative value here
                    # to trick it
                    set_timer(-61)
                    _RES2['messed'] = True
                    _RES2['fail'] = _RES2['succ'] = 0

                if session.worker_id > 100:
                    # starting to introduce errors passed the 100th
                    if random.randint(0, 10) == 1:
                        _RES2['fail'] += 1
                        raise AssertionError()
                    else:
                        _RES2['succ'] += 1

                # forces a switch
                await asyncio.sleep(0)

            stdout, stderr, rc = self._test_molotov('--sizing',
                                                    '--sizing-tolerance', '5',
                                                    '--console-update', '0',
                                                    '-cs', 'sizer',
                                                    'molotov.tests.test_run')

        ratio = float(_RES2['fail']) / float(_RES2['succ']) * 100.
        self.assertTrue(ratio < 20. and ratio > 5., ratio)

    @dedicatedloop
    def test_sizing_multiprocess_interrupted(self):

        counters = SharedCounters('OK', 'FAILED')

        @scenario()
        async def sizer(session):
            if random.randint(0, 10) == 1:
                counters['FAILED'] += 1
                raise AssertionError()
            else:
                counters['OK'] += 1

        async def _stop():
            await asyncio.sleep(2.)
            os.kill(os.getpid(), signal.SIGINT)

        asyncio.ensure_future(_stop())
        stdout, stderr, rc = self._test_molotov('--sizing', '-p', '3',
                                                '--sizing-tolerance', '90',
                                                '--console-update', '0',
                                                '-s', 'sizer',
                                                'molotov.tests.test_run')
        self.assertTrue("Sizing was not finished" in stdout)

    @dedicatedloop
    def test_use_extension(self):
        ext = os.path.join(_HERE, 'example5.py')

        @scenario(weight=10)
        async def simpletest(session):
            async with session.get('http://localhost:8888') as resp:
                assert resp.status == 200

        with coserver():
            stdout, stderr, rc = self._test_molotov('-cx', '--max-runs', '1',
                                                    '--use-extension=' + ext,
                                                    '-s',
                                                    'simpletest',
                                                    'molotov.tests.test_run')
        self.assertTrue("=>" in stdout)
        self.assertTrue("<=" in stdout)

    @dedicatedloop
    def test_use_extension_fail(self):
        ext = os.path.join(_HERE, 'exampleIDONTEXIST.py')

        @scenario(weight=10)
        async def simpletest(session):
            async with session.get('http://localhost:8888') as resp:
                assert resp.status == 200

        with coserver():
            stdout, stderr, rc = self._test_molotov('-cx', '--max-runs', '1',
                                                    '--use-extension=' + ext,
                                                    '-s',
                                                    'simpletest',
                                                    'molotov.tests.test_run')
        self.assertTrue("Cannot import" in stdout)

    @dedicatedloop
    def test_use_extension_module_name(self):
        ext = 'molotov.tests.example5'

        @scenario(weight=10)
        async def simpletest(session):
            async with session.get('http://localhost:8888') as resp:
                assert resp.status == 200

        with coserver():
            stdout, stderr, rc = self._test_molotov('-cx', '--max-runs', '1',
                                                    '--use-extension=' + ext,
                                                    '-s', 'simpletest',
                                                    'molotov.tests.test_run')
        self.assertTrue("=>" in stdout)
        self.assertTrue("<=" in stdout)

    @dedicatedloop
    def test_use_extension_module_name_fail(self):
        ext = 'IDONTEXTSIST'

        @scenario(weight=10)
        async def simpletest(session):
            async with session.get('http://localhost:8888') as resp:
                assert resp.status == 200

        with coserver():
            stdout, stderr, rc = self._test_molotov('-cx', '--max-runs', '1',
                                                    '--use-extension=' + ext,
                                                    '-s', 'simpletest',
                                                    'molotov.tests.test_run')
        self.assertTrue("Cannot import" in stdout)

    @dedicatedloop
    def test_quiet(self):

        @scenario(weight=10)
        async def here_three(session):
            _RES.append(3)

        stdout, stderr, rc = self._test_molotov('-cx', '--max-runs', '1', '-q',
                                                '-s', 'here_three',
                                                'molotov.tests.test_run')
        self.assertEqual(stdout, '')
        self.assertEqual(stderr, '')

    @dedicatedloop_noclose
    def test_slow_server_force_shutdown(self):
        @scenario(weight=10)
        async def _one(session):
            async with session.get('http://localhost:8888/slow') as resp:
                assert resp.status == 200
                _RES.append(1)

        args = self._get_args()
        args.duration = 2
        args.verbose = 2
        args.max_runs = 1
        args.force_shutdown = True
        start = time.time()
        with coserver():
            run(args)

        # makes sure the test is stopped even if the server
        # hangs a socket
        self.assertTrue(time.time() - start < 4)
        self.assertTrue(len(_RES) == 0)

    @dedicatedloop_noclose
    def test_slow_server_graceful(self):
        @scenario(weight=10)
        async def _one(session):
            async with session.get('http://localhost:8888/slow') as resp:
                assert resp.status == 200
                _RES.append(1)

        args = self._get_args()
        args.duration = 2
        args.verbose = 2
        args.max_runs = 1
        # graceful shutdown on the other hand will wait
        # for the worker completion
        args.graceful_shutdown = True

        start = time.time()
        with coserver():
            run(args)

        # makes sure the test finishes
        self.assertTrue(time.time() - start > 5)
        self.assertTrue(len(_RES) == 1)



"""

This Molotov script uses events to generate a success/failure output

"""
import json
import molotov
import time

_T = {}


def _now():
    return time.time() * 1000


@molotov.events()
async def record_time(event, **info):
    if event == 'scenario_start':
        scenario = info['scenario']
        index = (info['wid'], scenario['name'])
        _T[index] = _now()
    if event == 'scenario_success':
        scenario = info['scenario']
        index = (info['wid'], scenario['name'])
        start_time = _T.pop(index, None)
        duration = int(_now() - start_time)
        if start_time is not None:
            print(json.dumps({
                "ts": time.time(),
                "type": "scenario_success",
                "name": scenario['name'],
                "duration": duration,
            }))
    elif event == 'scenario_failure':
        scenario = info['scenario']
        exception = info['exception']
        index = (info['wid'], scenario['name'])
        start_time = _T.pop(index, None)
        duration = int(_now() - start_time)
        if start_time is not None:
            print(json.dumps({
                "ts": time.time(),
                "type": "scenario_failure",
                "name": scenario['name'],
                "exception": exception.__class__.__name__,
                "errorMessage": str(exception),
                "duration": duration
            }))



from molotov.listeners import BaseListener, EventSender
from molotov.tests.support import TestLoop, async_test, serialize


class TestListeners(TestLoop):
    @async_test
    async def test_add_listener(self, loop, console, results):
        class MyListener(BaseListener):
            def __init__(self):
                self.fired = False
                self.value = None

            def on_my_event(self, **options):
                self.fired = True
                self.value = options["value"]

        listener = MyListener()
        eventer = EventSender(console)
        eventer.add_listener(listener)
        await eventer.send_event("my_event", value=42)
        await serialize(console)

        self.assertTrue(listener.fired)
        self.assertEqual(listener.value, 42)

    @async_test
    async def test_buggy_listener(self, loop, console, results):
        class MyListener(BaseListener):
            def on_my_event(self, **options):
                raise Exception("Bam")

        listener = MyListener()
        eventer = EventSender(console)
        eventer.add_listener(listener)
        await eventer.send_event("my_event")
        resp = await serialize(console)
        self.assertTrue("Bam" in resp)



try:
    from molotov.api import (scenario, setup, global_setup, teardown,  # NOQA
                             global_teardown, setup_session,           # NOQA
                             teardown_session, scenario_picker,        # NOQA
                             events)                                   # NOQA
    from molotov.util import request, json_request                     # NOQA
    from molotov.util import set_var, get_var                          # NOQA
except ImportError:
    pass   # first import

__version__ = '1.6'



import socket
from urllib.parse import urlparse
import asyncio
from aiohttp.client import ClientSession, ClientRequest, ClientResponse
from aiohttp import TCPConnector

from molotov.util import resolve, IS_AIOHTTP2
from molotov.listeners import StdoutListener, EventSender


_HOST = socket.gethostname()


class LoggedClientRequest(ClientRequest):
    """Printable Request.
    """
    session = None

    if IS_AIOHTTP2:
        def send(self, *args, **kw):
            if self.session:
                event = self.session.send_event('sending_request',
                                                request=self)
                asyncio.ensure_future(event)
            response = super(LoggedClientRequest, self).send(*args, **kw)
            response.request = self
            return response
    else:
        async def send(self, *args, **kw):
            if self.session:
                event = self.session.send_event('sending_request',
                                                request=self)
                asyncio.ensure_future(event)
            response = await super(LoggedClientRequest, self).send(*args, **kw)
            response.request = self
            return response


class LoggedClientResponse(ClientResponse):
    request = None


class LoggedClientSession(ClientSession):
    """Session with printable requests and responses.
    """
    def __init__(self, loop, console, verbose=0, statsd=None, **kw):
        connector = kw.pop('connector', None)
        if connector is None:
            connector = TCPConnector(loop=loop, limit=None)
        super(LoggedClientSession,
              self).__init__(loop=loop,
                             request_class=LoggedClientRequest,
                             response_class=LoggedClientResponse,
                             connector=connector,  **kw)
        self.console = console
        self.request_class = LoggedClientRequest
        self.request_class.verbose = verbose
        self.verbose = verbose
        self.request_class.session = self
        self.request_class.response_class = LoggedClientResponse
        self.statsd = statsd
        self.eventer = EventSender(console,
                                   [StdoutListener(verbose=self.verbose,
                                                   console=self.console)])

    async def send_event(self, event, **options):
        await self.eventer.send_event(event, session=self, **options)

    def _dns_lookup(self, url):
        return resolve(url)[0]

    async def _request(self, *args, **kw):
        args = list(args)
        args[1] = self._dns_lookup(args[1])
        args = tuple(args)
        req = super(LoggedClientSession, self)._request

        if self.statsd:
            prefix = 'molotov.%(hostname)s.%(method)s.%(host)s.%(path)s'
            meth, url = args[:2]
            url = urlparse(url)
            path = url.path != '' and url.path or '/'

            data = {'method': meth,
                    'hostname': _HOST,
                    'host': url.netloc.split(":")[0],
                    'path': path}

            label = prefix % data

            @self.statsd.timer(label)
            async def request():
                resp = await req(*args, **kw)
                self.statsd.incr(label + '.' + str(resp.status))
                return resp

            resp = await request()
        else:
            resp = await req(*args, **kw)

        await self.send_event('response_received',
                              response=resp,
                              request=resp.request)
        return resp



import os
import sys
import argparse
import platform

from importlib import import_module
from importlib.util import spec_from_file_location, module_from_spec

from molotov.runner import Runner
from molotov.api import get_scenarios, get_scenario
from molotov import __version__
from molotov.util import expand_options, OptionError, printable_error
from molotov.sharedconsole import SharedConsole


PYPY = platform.python_implementation() == 'PyPy'


def _parser():
    parser = argparse.ArgumentParser(description='Load test.')

    parser.add_argument('scenario', default="loadtest.py",
                        help="path or module name that contains scenarii",
                        nargs="?")

    parser.add_argument('-s', '--single-mode', default=None, type=str,
                        help="Name of a single scenario to run once.")

    parser.add_argument('--config', default=None, type=str,
                        help='Point to a JSON config file.')

    parser.add_argument('--version', action='store_true', default=False,
                        help='Displays version and exits.')

    parser.add_argument('--debug', action='store_true', default=False,
                        help='Run the event loop in debug mode.')

    parser.add_argument('-v', '--verbose', action='count', default=0,
                        help=('Verbosity level. -v will display '
                              'tracebacks. -vv requests and responses.'))

    parser.add_argument('-w', '--workers', help='Number of workers',
                        type=int, default=1)

    parser.add_argument('--ramp-up', help='Ramp-up time in seconds',
                        type=float, default=0.)

    parser.add_argument('--sizing', help='Autosizing', action='store_true',
                        default=False)

    parser.add_argument('--sizing-tolerance', help='Sizing tolerance',
                        type=float, default=5.)

    parser.add_argument('--delay', help='Delay between each worker run',
                        type=float, default=0.)

    parser.add_argument('--console-update',
                        help='Delay between each console update',
                        type=float, default=0.2)

    parser.add_argument('-p', '--processes', help='Number of processes',
                        type=int, default=1)

    parser.add_argument('-d', '--duration', help='Duration in seconds',
                        type=int, default=86400)

    parser.add_argument('-r', '--max-runs', help='Maximum runs per worker',
                        type=int, default=None)

    parser.add_argument('-q', '--quiet', action='store_true', default=False,
                        help='Quiet')

    parser.add_argument('-x', '--exception', action='store_true',
                        default=False,
                        help='Stop on first failure.')

    parser.add_argument('-f', '--fail', type=int, default=None,
                        help='Number of failures required to fail')

    parser.add_argument('-c', '--console', action='store_true',
                        default=True,
                        help='Use simple console for feedback')

    parser.add_argument('--statsd', help='Activates statsd',
                        action='store_true', default=False)

    parser.add_argument('--statsd-address', help='Statsd Address',
                        type=str, default="udp://127.0.0.1:8125")

    parser.add_argument('--uvloop', help='Use uvloop', default=False,
                        action='store_true')

    parser.add_argument('--use-extension',
                        help='Imports a module containing Molotov extensions',
                        default=None, type=str, nargs='+')

    parser.add_argument('--force-shutdown',
                        help='Cancel all pending workers on shutdown',
                        default=False, action='store_true')

    return parser


def main(args=None):
    if args is None:
        parser = _parser()
        args = parser.parse_args()

    if args.version:
        print(__version__)
        sys.exit(0)

    if args.config:
        if args.scenario == 'loadtest.py':
            args.scenario = 'test'

        try:
            expand_options(args.config, args.scenario, args)
        except OptionError as e:
            print(str(e))
            sys.exit(0)

    if args.uvloop:
        if PYPY:
            print("You can't use uvloop with PyPy")     # pragma: no cover
            sys.exit(0)                                 # pragma: no cover

        try:
            import uvloop
        except ImportError:
            print('You need to install uvloop when using --uvloop')
            sys.exit(0)

        import asyncio
        asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())

    if args.sizing:
        # sizing is just ramping up workers indefinitely until
        # something things break. If the user has not set the values,
        # we do it here with 5 minutes and 500 workers
        if args.ramp_up == 0.:
            args.ramp_up = 300
        if args.workers == 1:
            args.workers = 500

    return run(args)


_SIZING = """\

Sizing is over!

Error Ratio %(RATIO).2f %% obtained with %(WORKER)d workers.

OVERALL: SUCCESSES: %(OK)d | FAILURES: %(FAILED)d
LAST MINUTE: SUCCESSES: %(MINUTE_OK)d | FAILURES: %(MINUTE_FAILED)d
"""

HELLO = '**** Molotov v%s. Happy breaking! ****' % __version__


def run(args):
    args.shared_console = SharedConsole(interval=args.console_update)

    if not args.quiet:
        print(HELLO)

    if args.use_extension:
        for extension in args.use_extension:
            if not args.quiet:
                print("Loading extension %r" % extension)
            if os.path.exists(extension):
                spec = spec_from_file_location("extension", extension)
                module = module_from_spec(spec)
                spec.loader.exec_module(module)
            else:
                try:
                    import_module(extension)
                except (ImportError, ValueError) as e:
                    print('Cannot import %r' % extension)
                    print('\n'.join(printable_error(e)))
                    sys.exit(1)

    if os.path.exists(args.scenario):
        spec = spec_from_file_location("loadtest", args.scenario)
        module = module_from_spec(spec)
        spec.loader.exec_module(module)
    else:
        try:
            import_module(args.scenario)
        except (ImportError, ValueError) as e:
            print('Cannot import %r' % args.scenario)
            print('\n'.join(printable_error(e)))
            sys.exit(1)

    if len(get_scenarios()) == 0:
        print('You need at least one scenario. No scenario was found.')
        print('A scenario with a weight of 0 is ignored')
        sys.exit(1)

    if args.verbose > 0 and args.quiet:
        print("You can't use -q and -v at the same time")
        sys.exit(1)

    if args.single_mode:
        if get_scenario(args.single_mode) is None:
            print("Can't find %r in registered scenarii" % args.single_mode)
            sys.exit(1)

    res = Runner(args)()

    def _dict(counters):
        res = {}
        for k, v in counters.items():
            if k == 'RATIO':
                res[k] = float(v.value) / 100.
            else:
                res[k] = v.value
        return res

    res = _dict(res)

    if not args.quiet:
        if args.sizing:
            if res['REACHED'] == 1:
                print(_SIZING % res)
            else:
                print('Sizing was not finished. (interrupted)')
        else:
            print('SUCCESSES: %(OK)d | FAILURES: %(FAILED)d\r' % res)
        print('*** Bye ***')
        if args.fail is not None and res['FAILED'] >= args.fail:
            sys.exit(1)



import asyncio
import time

from molotov.listeners import EventSender
from molotov.session import LoggedClientSession as Session
from molotov.api import get_fixture, pick_scenario, get_scenario
from molotov.util import (cancellable_sleep, is_stopped, set_timer, get_timer,
                          stop)


class FixtureError(Exception):
    pass


def _now():
    return int(time.time())


class Worker(object):
    """"The Worker class creates a Session and runs scenario.
    """
    def __init__(self, wid, results, console, args, statsd=None,
                 delay=0, loop=None):
        self.wid = wid
        self.results = results
        self.console = console
        self.loop = loop or asyncio.get_event_loop()
        self.args = args
        self.statsd = statsd
        self.delay = delay
        self.count = 0
        self.worker_start = 0
        self.eventer = EventSender(console)
        # fixtures
        self._session_setup = get_fixture('setup_session')
        self._session_teardown = get_fixture('teardown_session')
        self._setup = get_fixture('setup')
        self._teardown = get_fixture('teardown')

    async def send_event(self, event, **options):
        await self.eventer.send_event(event, wid=self.wid, **options)

    async def run(self):
        if self.delay > 0.:
            await cancellable_sleep(self.delay)
        if is_stopped():
            return
        self.results['WORKER'] += 1
        try:
            res = await self._run()
        finally:
            self.teardown()
            self.results['WORKER'] -= 1
        return res

    def _may_run(self):
        if is_stopped():
            return False
        if _now() - self.worker_start > self.args.duration:
            return False
        if self.results['REACHED'] == 1:
            return False
        if self.args.max_runs and self.count > self.args.max_runs:
            return False
        return True

    async def setup(self):
        if self._setup is None:
            return {}
        try:
            options = await self._setup(self.wid, self.args)
        except Exception as e:
            self.console.print_error(e)
            raise FixtureError(str(e))

        if options is None:
            options = {}
        elif not isinstance(options, dict):
            msg = 'The setup function needs to return a dict'
            self.console.print(msg)
            raise FixtureError(msg)

        return options

    async def session_setup(self, session):
        if self._session_setup is None:
            return
        try:
            await self._session_setup(self.wid, session)
        except Exception as e:
            self.console.print_error(e)
            raise FixtureError(str(e))

    async def session_teardown(self, session):
        if self._session_teardown is None:
            return
        try:
            await self._session_teardown(self.wid, session)
        except Exception as e:
            # we can't stop the teardown process
            self.console.print_error(e)

    async def _run(self):
        verbose = self.args.verbose
        exception = self.args.exception

        if self.args.single_mode:
            single = get_scenario(self.args.single_mode)
        else:
            single = None

        self.count = 1
        self.worker_start = _now()

        try:
            options = await self.setup()
        except FixtureError:
            stop()
            return

        async with Session(self.loop, self.console, verbose, self.statsd,
                           **options) as session:
            session.args = self.args
            session.worker_id = self.wid

            try:
                await self.session_setup(session)
            except FixtureError:
                stop()
                return

            while self._may_run():
                step_start = _now()
                session.step = self.count
                result = await self.step(self.count, session, scenario=single)
                if result == 1:
                    self.results['OK'] += 1
                    self.results['MINUTE_OK'] += 1
                elif result == -1:
                    self.results['FAILED'] += 1
                    self.results['MINUTE_FAILED'] += 1
                    if exception:
                        stop()

                if not is_stopped() and self._reached_tolerance(step_start):
                    stop()
                    cancellable_sleep.cancel_all()
                    break

                self.count += 1
                if self.args.delay > 0.:
                    await cancellable_sleep(self.args.delay)
                else:
                    # forces a context switch
                    await asyncio.sleep(0)

            await self.session_teardown(session)

    def teardown(self):
        if self._teardown is None:
            return
        try:
            self._teardown(self.wid)
        except Exception as e:
            # we can't stop the teardown process
            self.console.print_error(e)

    def _reached_tolerance(self, current_time):
        if not self.args.sizing:
            return False

        if current_time - get_timer() > 60:
            # we need to reset the tolerance counters
            set_timer(current_time)
            self.results['MINUTE_OK'].value = 0
            self.results['MINUTE_FAILED'].value = 0
            return False

        OK = self.results['MINUTE_OK'].value
        FAILED = self.results['MINUTE_FAILED'].value

        if OK + FAILED < 100:
            # we don't have enough samples
            return False

        current_ratio = float(FAILED) / float(OK) * 100.
        reached = current_ratio > self.args.sizing_tolerance
        if reached:
            self.results['REACHED'].value = 1
            self.results['RATIO'].value = int(current_ratio * 100)

        return reached

    async def step(self, step_id, session, scenario=None):
        """ single scenario call.

        When it returns 1, it works. -1 the script failed,
        0 the test is stopping or needs to stop.
        """
        if scenario is None:
            scenario = pick_scenario(self.wid, step_id)
        try:
            await self.send_event('scenario_start', scenario=scenario)

            await scenario['func'](session, *scenario['args'],
                                   **scenario['kw'])

            await self.send_event('scenario_success', scenario=scenario)

            if scenario['delay'] > 0.:
                await cancellable_sleep(scenario['delay'])
            return 1
        except Exception as exc:
            await self.send_event('scenario_failure',
                                  scenario=scenario,
                                  exception=exc)
            if self.args.verbose > 0:
                self.console.print_error(exc)
                await self.console.flush()

        return -1



import os
import pytest
from molotov import __version__
from molotov.slave import main
from molotov.tests.support import TestLoop, dedicatedloop, set_args


_REPO = 'https://github.com/loads/molotov'
NO_INTERNET = os.environ.get("NO_INTERNET") is not None


@pytest.mark.skipif(NO_INTERNET, reason="This test requires internet access")
class TestSlave(TestLoop):
    @dedicatedloop
    def test_main(self):
        with set_args('moloslave', _REPO, 'test') as out:
            main()
        if os.environ.get("TRAVIS") is not None:
            return
        output = out[0].read()
        self.assertTrue('Preparing 1 worker...' in output, output)
        self.assertTrue('OK' in output, output)

    @dedicatedloop
    def test_fail(self):
        with set_args('moloslave', _REPO, 'fail'):
            self.assertRaises(Exception, main)

    @dedicatedloop
    def test_version(self):
        with set_args('moloslave', '--version') as out:
            try:
                main()
            except SystemExit:
                pass
        version = out[0].read().strip()
        self.assertTrue(version, __version__)



import io

from molotov.api import get_fixture

_UNREADABLE = "***WARNING: Molotov can't display this body***"
_BINARY = "**** Binary content ****"
_FILE = "**** File content ****"
_COMPRESSED = ('gzip', 'compress', 'deflate', 'identity', 'br')


class BaseListener(object):
    async def __call__(self, event, **options):
        attr = getattr(self, 'on_' + event, None)
        if attr is not None:
            await attr(**options)


class StdoutListener(BaseListener):
    def __init__(self, **options):
        self.verbose = options.get('verbose', 0)
        self.console = options['console']

    def _body2str(self, body):
        try:
            from aiohttp.payload import Payload
        except ImportError:
            Payload = None

        if Payload is not None and isinstance(body, Payload):
            body = body._value

        if isinstance(body, io.IOBase):
            return _FILE

        if not isinstance(body, str):
            try:
                body = str(body, 'utf8')
            except UnicodeDecodeError:
                return _UNREADABLE

        return body

    async def on_sending_request(self, session, request):
        if self.verbose < 2:
            return
        raw = '>' * 45
        raw += '\n' + request.method + ' ' + str(request.url)
        if len(request.headers) > 0:
            headers = '\n'.join('%s: %s' % (k, v) for k, v in
                                request.headers.items())
            raw += '\n' + headers
        if request.headers.get('Content-Encoding') in _COMPRESSED:
            raw += '\n\n' + _BINARY + '\n'
        elif request.body:
            raw += '\n\n' + self._body2str(request.body) + '\n'

        self.console.print(raw)

    async def on_response_received(self, session, response, request):
        if self.verbose < 2:
            return
        raw = '\n' + '=' * 45 + '\n'
        raw += 'HTTP/1.1 %d %s\n' % (response.status, response.reason)
        items = response.headers.items()
        headers = '\n'.join('{}: {}'.format(k, v) for k, v in items)
        raw += headers
        if response.headers.get('Content-Encoding') in _COMPRESSED:
            raw += '\n\n' + _BINARY
        elif response.content:
            content = await response.content.read()
            if len(content) > 0:
                # put back the data in the content
                response.content.unread_data(content)
                try:
                    raw += '\n\n' + content.decode()
                except UnicodeDecodeError:
                    raw += '\n\n' + _UNREADABLE
            else:
                raw += '\n\n'

        raw += '\n' + '<' * 45 + '\n'
        self.console.print(raw)


class CustomListener(object):
    def __init__(self, fixture):
        self.fixture = fixture

    async def __call__(self, event, **options):
        await self.fixture(event, **options)


class EventSender(object):
    def __init__(self, console, listeners=None):
        self.console = console
        if listeners is None:
            listeners = []
        self._listeners = listeners
        self._stopped = False

        fixture_listeners = get_fixture('events')
        if fixture_listeners is not None:
            for listener in fixture_listeners:
                self.add_listener(CustomListener(listener))

    def add_listener(self, listener):
        self._listeners.append(listener)

    async def stop(self):
        self._stopped = True

    def stopped(self):
        return self._stopped

    async def send_event(self, event, **options):
        for listener in self._listeners:
            try:
                await listener(event, **options)
            except Exception as e:
                self.console.print_error(e)



import multiprocessing


class SharedCounter(object):
    """A multi-process compatible counter.
    """
    def __init__(self, name):
        self._val = multiprocessing.Value('i', 0)
        self._name = name

    def __eq__(self, other):
        return self.__cmp__(other) == 0

    def __ne__(self, other):
        return self.__cmp__(other) != 0

    def __gt__(self, other):
        return self.__cmp__(other) > 0

    def __ge__(self, other):
        return self.__cmp__(other) >= 0

    def __lt__(self, other):
        return self.__cmp__(other) < 0

    def __le__(self, other):
        return self.__cmp__(other) <= 0

    def __cmp__(self, other):
        if isinstance(other, SharedCounter):
            other = other.value
        if not isinstance(other, int):
            raise TypeError(other)
        if self._val.value == other:
            return 0
        elif self._val.value > other:
            return 1
        return -1

    def __repr__(self):
        return '<SharedCounter %d>' % self._val.value

    def __iadd__(self, other):
        self.__add__(other)
        return self

    def __isub__(self, other):
        self.__sub__(other)
        return self

    def __add__(self, other):
        with self._val.get_lock():
            if isinstance(other, SharedCounter):
                other = other.value
            if not isinstance(other, int):
                raise NotImplementedError()
            self._val.value += other

    def __sub__(self, other):
        self.__add__(-other)

    @property
    def value(self):
        return self._val.value

    @value.setter
    def value(self, _value):
        with self._val.get_lock():
            if isinstance(_value, SharedCounter):
                _value = _value.value
            if not isinstance(_value, int):
                raise TypeError(_value)
            self._val.value = _value


class SharedCounters(object):
    """Mapping of SharedCounter items.
    """
    def __init__(self, *keys):
        self._counters = {}
        for key in keys:
            self._counters[key] = SharedCounter(key)

    def items(self):
        return self._counters.items()

    def values(self):
        return self._counters.values()

    def __iter__(self):
        return self._counters.__iter__()

    def keys(self):
        return self._counters.keys()

    def __contains__(self, key):
        return key in self._counters

    def __repr__(self):
        return repr(self._counters)

    def __setitem__(self, key, value):
        if key not in self._counters:
            raise KeyError(key)
        self._counters[key].value = value

    def __getitem__(self, key):
        return self._counters[key]



import unittest
import multiprocessing
from molotov.sharedcounter import SharedCounters, SharedCounter


# pre-forked variable
_DATA = SharedCounters('test')


def run_worker(value):
    _DATA['test'] += value
    _DATA['test'] -= value
    _DATA['test'] += value


class TestSharedCounters(unittest.TestCase):
    def test_operators(self):
        c1 = SharedCounter('ok')
        c2 = SharedCounter('ok')
        c1.value = 4
        c2.value = 5
        self.assertTrue(c1 <= c2)
        self.assertTrue(c1 < c2)
        self.assertTrue(c1 >= 2)
        self.assertTrue(c1 > 2)
        self.assertTrue(c1 == 4)
        self.assertTrue(c1 != 5)
        c2.value = 4
        c2 += SharedCounter('ok')
        self.assertTrue(c1 == c2)
        repr(c1)
        str(c1)

        def _t():
            c = SharedCounter('ok')
            c += 6.2

        self.assertRaises(NotImplementedError, _t)

        def _c():
            SharedCounter('ok') != 6.3

        self.assertRaises(TypeError, _c)

    def test_interface(self):
        data = SharedCounters('one', 'two')
        self.assertTrue('one' in data)
        self.assertEqual(len(data.keys()), 2)

        for key in data:
            data[key] = 0
            self.assertTrue(data[key], 0)

        for key, value in data.items():
            data[key] = value
            self.assertTrue(data[key], value)

        data.values()
        repr(data)
        str(data)
        self.assertRaises(KeyError, data.__setitem__, 'meh', 1)
        self.assertRaises(TypeError, data.__setitem__, 'one', '1')

    def test_mapping(self):
        # making sure it works like a defaultdict(int)
        data = SharedCounters('one', 'two')
        self.assertTrue(data['one'].value == 0)
        data['one'] += 10
        data['one'] -= 1
        data['two'] = 4
        self.assertTrue(data['one'].value, 9)
        self.assertTrue(data['two'].value, 4)

    def test_multiprocess(self):
        # now let's try with several processes
        pool = multiprocessing.Pool(10)
        try:
            inputs = [1] * 3000
            pool.map(run_worker, inputs)
            self.assertEqual(_DATA['test'].value, 3000)
        finally:
            pool.close()



import gzip
from aiohttp.client_reqrep import ClientRequest
from yarl import URL
from unittest.mock import patch

from molotov.listeners import BaseListener
import molotov.session
from molotov.tests.support import coserver, Response, Request
from molotov.tests.support import TestLoop, async_test, serialize


class TestLoggedClientSession(TestLoop):

    def _get_session(self, *args, **kw):
        return molotov.session.LoggedClientSession(*args, **kw)

    @async_test
    async def test_add_listener(self, loop, console, results):
        class MyListener(BaseListener):
            def __init__(self):
                self.responses = []

            def on_response_received(self, **options):
                self.responses.append(options['response'])

        lis = MyListener()
        async with self._get_session(loop, console,
                                     verbose=2) as session:
            session.eventer.add_listener(lis)
            request = Request()
            binary_body = b''
            response = Response(body=binary_body)
            await session.send_event('response_received',
                                     response=response,
                                     request=request)

        await serialize(console)
        self.assertEqual(lis.responses, [response])

    @async_test
    async def test_empty_response(self, loop, console, results):
        async with self._get_session(loop, console,
                                     verbose=2) as session:
            request = Request()
            binary_body = b''
            response = Response(body=binary_body)
            await session.send_event('response_received',
                                     response=response,
                                     request=request)

        await serialize(console)

    @async_test
    async def test_encoding(self, loop, console, results):
        async with self._get_session(loop, console,
                                     verbose=2) as session:
            request = Request()
            binary_body = b'MZ\x90\x00\x03\x00\x00\x00\x04\x00'
            response = Response(body=binary_body)
            await session.send_event('response_received',
                                     response=response,
                                     request=request)

        res = await serialize(console)
        wanted = "can't display this body"
        self.assertTrue(wanted in res)

    @async_test
    async def test_request(self, loop, console, results):
        with coserver():
            async with self._get_session(loop, console,
                                         verbose=2) as session:
                async with session.get('http://localhost:8888') as resp:
                    self.assertEqual(resp.status, 200)

            res = await serialize(console)
            self.assertTrue('GET http://127.0.0.1:8888' in res)

    @async_test
    async def test_not_verbose(self, loop, console, results):
        async with self._get_session(loop, console,
                                     verbose=1) as session:
            req = ClientRequest('GET', URL('http://example.com'))
            await session.send_event('sending_request', request=req)

            response = Response(body='')
            request = Request()
            await session.send_event('response_received',
                                     response=response,
                                     request=request)

        res = await serialize(console)
        self.assertEqual(res, '')

    @async_test
    async def test_gzipped_request(self, loop, console, results):
        async with self._get_session(loop, console,
                                     verbose=2) as session:
            binary_body = gzip.compress(b'some gzipped data')
            req = ClientRequest('GET', URL('http://example.com'),
                                data=binary_body)
            req.headers['Content-Encoding'] = 'gzip'
            await session.send_event('sending_request', request=req)

        res = await serialize(console)
        self.assertTrue("Binary" in res, res)

    @async_test
    async def test_file_request(self, loop, console, results):
        async with self._get_session(loop, console,
                                     verbose=2) as session:
            with open(__file__) as f:
                req = ClientRequest('POST', URL('http://example.com'),
                                    data=f)
                req.headers['Content-Encoding'] = 'something/bin'
                await session.send_event('sending_request', request=req)

        res = await serialize(console)
        self.assertTrue("File" in res, res)

    @async_test
    async def test_binary_file_request(self, loop, console, results):
        async with self._get_session(loop, console,
                                     verbose=2) as session:
            with open(__file__, 'rb') as f:
                req = ClientRequest('POST', URL('http://example.com'),
                                    data=f)
                req.headers['Content-Encoding'] = 'something/bin'
                await session.send_event('sending_request', request=req)

        res = await serialize(console)
        self.assertTrue("File" in res, res)

    @async_test
    async def test_gzipped_response(self, loop, console, results):
        async with self._get_session(loop, console,
                                     verbose=2) as session:
            request = Request()
            binary_body = gzip.compress(b'some gzipped data')
            response = Response(body=binary_body)
            response.headers['Content-Encoding'] = 'gzip'
            await session.send_event('response_received',
                                     response=response,
                                     request=request)

        res = await serialize(console)
        self.assertTrue("Binary" in res, res)

    @async_test
    async def test_cantread_request(self, loop, console, results):
        async with self._get_session(loop, console,
                                     verbose=2) as session:
            binary_body = gzip.compress(b'some gzipped data')
            req = ClientRequest('GET', URL('http://example.com'),
                                data=binary_body)
            await session.send_event('sending_request', request=req)

        res = await serialize(console)
        self.assertTrue("display this body" in res, res)

    @async_test
    async def test_old_request_version(self, loop, console, results):

        orig_import = __import__

        def import_mock(name, *args, **kw):
            if name == 'aiohttp.payload':
                raise ImportError()
            return orig_import(name, *args, **kw)

        with patch('builtins.__import__', side_effect=import_mock):
            async with self._get_session(loop, console,
                                         verbose=2) as session:
                body = "ok man"
                req = ClientRequest('GET', URL('http://example.com'),
                                    data=body)
                req.body = req.body._value
                await session.send_event('sending_request', request=req)

        res = await serialize(console)
        self.assertTrue("ok man" in res, res)



import sys
import asyncio
import multiprocessing
import os
from queue import Empty

from molotov.util import cancellable_sleep, printable_error


class SharedConsole(object):
    """Multi-process compatible stdout console.
    """
    def __init__(self, interval=.1, max_lines_displayed=20):
        self._stream = multiprocessing.Queue()
        self._interval = interval
        self._stop = True
        self._creator = os.getpid()
        self._stop = False
        self._max_lines_displayed = max_lines_displayed

    async def stop(self):
        self._stop = True
        while True:
            try:
                sys.stdout.write(self._stream.get_nowait())
            except Empty:
                break
        sys.stdout.flush()

    async def flush(self):
        sys.stdout.flush()
        await asyncio.sleep(0)

    async def display(self):
        if os.getpid() != self._creator:
            return

        while not self._stop:
            lines_displayed = 0
            while True:
                try:
                    line = self._stream.get_nowait()
                    sys.stdout.write(line)
                    lines_displayed += 1
                except Empty:
                    break
                if self._stop or lines_displayed > self._max_lines_displayed:
                    break
                else:
                    await asyncio.sleep(0)
            sys.stdout.flush()
            if not self._stop:
                await cancellable_sleep(self._interval)

    def print(self, line, end='\n'):
        if os.getpid() != self._creator:
            line = "[%d] %s" % (os.getpid(), line)
        line += end
        self._stream.put_nowait(line)

    def print_error(self, error, tb=None):
        for line in printable_error(error, tb):
            self.print(line)

    def print_block(self, start, callable, end='OK'):
        if os.getpid() != self._creator:
            prefix = "[%d] " % os.getpid()
        else:
            prefix = ''
        self._stream.put(prefix + start + '...\n')
        res = callable()
        self._stream.put(prefix + 'OK\n')
        return res



"""

This Molotov script demonstrates how to hook events.

"""

import molotov


@molotov.events()
async def print_request(event, **info):
    if event == 'sending_request':
        print("=>")


@molotov.events()
async def print_response(event, **info):
    if event == 'response_received':
        print("<=")


@molotov.scenario(100)
async def scenario_one(session):
    async with session.get('http://localhost:8080') as resp:
        res = await resp.json()
        assert res['result'] == 'OK'
        assert resp.status == 200



from molotov.api import pick_scenario, scenario, get_scenarios, setup
from molotov.tests.support import TestLoop, async_test


class TestUtil(TestLoop):
    def test_pick_scenario(self):

        @scenario(weight=10)
        async def _one(self):
            pass

        @scenario(weight=90)
        async def _two(self):
            pass

        picked = [pick_scenario()['name'] for i in range(100)]
        ones = len([f for f in picked if f == '_one'])
        self.assertTrue(ones < 20)

    @async_test
    async def test_can_call(self, loop, console, results):
        @setup()
        async def _setup(self):
            pass

        @scenario(weight=10)
        async def _one(self):
            pass

        # can still be called
        await _one(self)

        # same for fixtures
        await _setup(self)

    def test_default_weight(self):
        @scenario()
        async def _default_weight(self):
            pass

        self.assertEqual(len(get_scenarios()), 1)
        self.assertEqual(get_scenarios()[0]['weight'], 1)

    def test_no_scenario(self):
        @scenario(weight=0)
        async def _one(self):
            pass

        @scenario(weight=0)
        async def _two(self):
            pass

        self.assertEqual(get_scenarios(), [])

    def test_scenario_not_coroutine(self):
        try:
            @scenario(weight=1)
            def _one(self):
                pass
        except TypeError:
            return
        raise AssertionError("Should raise")

    def test_setup_not_coroutine(self):
        try:
            @setup()
            def _setup(self):
                pass

            @scenario(weight=90)
            async def _two(self):
                pass
        except TypeError:
            return
        raise AssertionError("Should raise")

    def test_two_fixtures(self):
        try:
            @setup()
            async def _setup(self):
                pass

            @setup()
            async def _setup2(self):
                pass

            @scenario(weight=90)
            async def _two(self):
                pass
        except ValueError:
            return
        raise AssertionError("Should raise")



from contextlib import suppress
import signal
import multiprocessing
import asyncio
import os

from molotov.api import get_fixture
from molotov.listeners import EventSender
from molotov.stats import get_statsd_client
from molotov.sharedcounter import SharedCounters
from molotov.util import cancellable_sleep, stop, is_stopped, set_timer
from molotov.worker import Worker


class Runner(object):
    """Manages processes & workers and grabs results.
    """
    def __init__(self, args, loop=None):
        self.args = args
        self.console = self.args.shared_console
        if loop is None:
            loop = asyncio.get_event_loop()
        self.loop = loop
        # the stastd client gets initialized after we fork
        # processes in case -p was used
        self.statsd = None
        self._tasks = []
        self._procs = []
        self._results = SharedCounters('WORKER', 'REACHED', 'RATIO', 'OK',
                                       'FAILED', 'MINUTE_OK', 'MINUTE_FAILED')
        self.eventer = EventSender(self.console)

    def _set_statsd(self):
        if self.args.statsd:
            self.statsd = get_statsd_client(self.args.statsd_address,
                                            loop=self.loop)
        else:
            self.statsd = None

    def gather(self, *futures):
        return asyncio.gather(*futures, loop=self.loop, return_exceptions=True)

    def ensure_future(self, coro):
        return asyncio.ensure_future(coro, loop=self.loop)

    def __call__(self):
        global_setup = get_fixture('global_setup')
        if global_setup is not None:
            try:
                global_setup(self.args)
            except Exception as e:
                self.console.print("The global_setup() fixture failed")
                self.console.print_error(e)
                raise

        try:
            return self._launch_processes()
        finally:
            global_teardown = get_fixture('global_teardown')
            if global_teardown is not None:
                try:
                    global_teardown()
                except Exception as e:
                    # we can't stop the teardown process
                    self.console.print_error(e)

    def _launch_processes(self):
        args = self.args
        signal.signal(signal.SIGINT, self._shutdown)
        signal.signal(signal.SIGTERM, self._shutdown)
        args.original_pid = os.getpid()

        if args.processes > 1:
            if not args.quiet:
                self.console.print('Forking %d processes' % args.processes)
            jobs = []
            for i in range(args.processes):
                p = multiprocessing.Process(target=self._process)
                jobs.append(p)
                p.start()

            for job in jobs:
                self._procs.append(job)

            async def run(quiet, console):
                while len(self._procs) > 0:
                    if not quiet:
                        console.print(self.display_results(), end='\r')
                    for job in jobs:
                        if job.exitcode is not None and job in self._procs:
                            self._procs.remove(job)
                    await cancellable_sleep(args.console_update)
                await self.console.stop()
                await self.eventer.stop()

            tasks = [self.ensure_future(self.console.display()),
                     self.ensure_future(self._send_workers_event(1)),
                     self.ensure_future(run(args.quiet, self.console))]
            self.loop.run_until_complete(self.gather(*tasks))
        else:
            self._process()

        return self._results

    def _shutdown(self, signal, frame):
        stop()
        self._kill_tasks()
        # send sigterms
        for proc in self._procs:
            proc.terminate()

    def _runner(self):
        args = self.args

        def _prepare():
            tasks = []
            delay = 0
            if args.ramp_up > 0.:
                step = args.ramp_up / args.workers
            else:
                step = 0.
            for i in range(self.args.workers):
                worker = Worker(i, self._results, self.console, self.args,
                                self.statsd, delay, self.loop)
                f = self.ensure_future(worker.run())
                tasks.append(f)
                delay += step
            return tasks

        if self.args.quiet:
            return _prepare()
        else:
            msg = 'Preparing {} worker{}'
            msg = msg.format(args.workers, 's' if args.workers > 1 else '')
            return self.console.print_block(msg, _prepare)

    def _process(self):
        set_timer()

        # coroutine that will kill everything when duration is up
        if self.args.duration and self.args.force_shutdown:
            async def _duration_killer():
                cancelled = object()
                res = await cancellable_sleep(self.args.duration,
                                              result=cancelled)
                if res is cancelled or (res and not res.canceled()):
                    self._shutdown(None, None)
                    await asyncio.sleep(0)

            _duration_killer = self.ensure_future(_duration_killer())
        else:
            _duration_killer = None

        if self.args.processes > 1:
            signal.signal(signal.SIGINT, self._shutdown)
            signal.signal(signal.SIGTERM, self._shutdown)
            self.loop = asyncio.new_event_loop()
            asyncio.set_event_loop(self.loop)

        if self.args.debug:
            self.console.print('**** RUNNING IN DEBUG MODE == SLOW ****')
            self.loop.set_debug(True)

        self._set_statsd()

        if self.args.original_pid == os.getpid():
            self._tasks.append(self.ensure_future(self._send_workers_event(1)))
            if not self.args.quiet:
                fut = self._display_results(self.args.console_update)
                update = self.ensure_future(fut)
                display = self.ensure_future(self.console.display())
                display = self.gather(update, display)
                self._tasks.append(display)

        workers = self.gather(*self._runner())

        def _stop(cb):
            if _duration_killer is not None:
                if not _duration_killer.done():
                    _duration_killer.cancel()
            stop()

        workers.add_done_callback(_stop)
        self._tasks.append(workers)

        try:
            self.loop.run_until_complete(self.gather(*self._tasks))
        finally:
            self._kill_tasks()
            if self.statsd is not None:
                self.statsd.close()
            self.loop.run_until_complete(self.ensure_future(asyncio.sleep(0)))
            self.loop.close()

    def _kill_tasks(self):
        cancellable_sleep.cancel_all()
        for task in reversed(self._tasks):
            with suppress(asyncio.CancelledError):
                task.cancel()
        for task in self._tasks:
            del task
        self._tasks[:] = []

    def display_results(self):
        ok, fail = self._results['OK'].value, self._results['FAILED'].value
        workers = self._results['WORKER'].value
        pat = 'SUCCESSES: %s | FAILURES: %s | WORKERS: %s'
        return pat % (ok, fail, workers)

    async def _display_results(self, update_interval):
        while not is_stopped():
            self.console.print(self.display_results(), end='\r')
            await cancellable_sleep(update_interval)
        await self.console.stop()

    async def _send_workers_event(self, update_interval):
        while not self.eventer.stopped() and not is_stopped():
            workers = self._results['WORKER'].value
            await self.eventer.send_event('current_workers', workers=workers)
            await cancellable_sleep(update_interval)
